@@ -43,15 +43,21 @@ function destroyer(stream, reading, writing, callback) {
 
     // request.destroy just do .end - .abort is what we want
     if (isRequest(stream)) return stream.abort();
-    if (typeof stream.destroy === 'function') return stream.destroy();
+    if (typeof stream.destroy === 'function') {
+      if (stream.req && stream._writableState === undefined) {
+        // This is a ClientRequest
+        // TODO(mcollina): backward compatible fix to avoid crashing.
+        // Possibly remove in a later semver-major change.
+        stream.req.on('error', noop);
+      }
+      return stream.destroy(err);
+    }
 
     callback(err || new ERR_STREAM_DESTROYED('pipe'));
   };
 }
 
-function call(fn) {
-  fn();
-}
+function noop() {}
 
 function pipe(from, to) {
   return from.pipe(to);
@@ -81,9 +87,15 @@ function pipeline(...streams) {
     const writing = i > 0;
     return destroyer(stream, reading, writing, function(err) {
       if (!error) error = err;
-      if (err) destroys.forEach(call);
+      if (err) {
+        for (const destroy of destroys) {
+          destroy(err);
+        }
+      }
       if (reading) return;
-      destroys.forEach(call);
+      for (const destroy of destroys) {
+        destroy();
+      }
       callback(error);
     });
   });
