@@ -82,8 +82,13 @@ struct napi_env__ {
     return v8::Just(true);
   }
 
-  template <typename T, typename U>
-  void CallIntoModule(T&& call, U&& handle_exception) {
+  static inline void
+  HandleThrow(napi_env env, v8::Local<v8::Value> value) {
+    env->isolate->ThrowException(value);
+  }
+
+  template <typename T, typename U = decltype(HandleThrow)>
+  inline void CallIntoModule(T&& call, U&& handle_exception = HandleThrow) {
     int open_handle_scopes_before = open_handle_scopes;
     int open_callback_scopes_before = open_callback_scopes;
     napi_clear_last_error(this);
@@ -96,13 +101,6 @@ struct napi_env__ {
     }
   }
 
-  template <typename T>
-  void CallIntoModuleThrow(T&& call) {
-    CallIntoModule(call, [&](napi_env env, v8::Local<v8::Value> value) {
-      env->isolate->ThrowException(value);
-    });
-  }
-
   v8impl::Persistent<v8::Value> last_exception;
 
   // We store references in two different lists, depending on whether they have
@@ -199,6 +197,17 @@ napi_status napi_set_last_error(napi_env env, napi_status error_code,
     }                                                              \
   } while (0)
 
+#define RETURN_STATUS_IF_FALSE_WITH_PREAMBLE(env, condition, status)           \
+  do {                                                                         \
+    if (!(condition)) {                                                        \
+      return napi_set_last_error(                                              \
+          (env), try_catch.HasCaught() ? napi_pending_exception : (status));   \
+    }                                                                          \
+  } while (0)
+
+#define CHECK_MAYBE_EMPTY_WITH_PREAMBLE(env, maybe, status)                    \
+  RETURN_STATUS_IF_FALSE_WITH_PREAMBLE((env), !((maybe).IsEmpty()), (status))
+
 namespace v8impl {
 
 //=== Conversion between V8 Handles and napi_value ========================
